# Generated by Django 3.1.1 on 2020-10-08 22:02

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='APIStat',
            fields=[
                ('record_time', models.DateTimeField(auto_now=True, primary_key=True, serialize=False)),
                ('method', models.CharField(max_length=10)),
                ('path', models.CharField(db_index=True, max_length=255)),
                ('query', models.CharField(max_length=2000)),
                ('ip', models.GenericIPAddressField()),
            ],
            options={
                'ordering': ['-record_time', 'path'],
                'get_latest_by': ['-record_time', 'path'],
            },
        ),
    ]



from django.contrib import admin
from django.db import models

from .models import APIStat


class APIStatAdmin(admin.ModelAdmin):
    date_hierarchy = 'record_time'
    list_display = ('record_time', 'method', 'domain', 'path', 'user', 'delay', 'status')
    readonly_fields = ('record_time', 'method', 'domain', 'path', 'delay', 'status',
                       'query', 'ip', 'total_count', 'domain_count', 'path_count',
                       'avg_delay', 'avg_domain_delay', 'avg_path_delay',
                       'max_delay', 'max_domain_delay', 'max_path_delay',)
    list_filter = ('method', 'domain', 'path', 'user')
    fieldsets = (
        (None, {
            'fields': ('record_time', 'domain', 'path', 'user')
        }),
        ('Request', {
            'fields': ('method', 'query', 'ip'),
        }),
        ('Response', {
            'fields': ('delay', 'status'),
        }),
        ('Stats', {
            'fields': ('total_count', 'max_delay', 'avg_delay',
                       'domain_count', 'max_domain_delay', 'avg_domain_delay',
                       'path_count', 'max_path_delay', 'avg_path_delay'),
        }),
    )


admin.site.register(APIStat, APIStatAdmin)



# Generated by Django 3.1.1 on 2020-10-08 22:03

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('apistats', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='apistat',
            name='ip',
            field=models.GenericIPAddressField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='apistat',
            name='query',
            field=models.CharField(blank=True, max_length=2000),
        ),
    ]



from django.contrib.auth.models import User, AnonymousUser
from django.test import RequestFactory,TestCase

from .models import APIStat


class APIStatsTestCase(TestCase):

    def setUp(self):
        self.factory = RequestFactory()
        self.user = User.objects.create_user(
            username='jacob', email='jacob@…', password='top_secret')

    def test_create(self):
        request = self.factory.get('/test/a/path/')
        request.user = AnonymousUser()
        delay = 100
        status = 200
        apistat = APIStat.objects.create_from_request(request=request, delay=delay, status=status)
        self.assertIsNotNone(apistat)
        self.assertEqual(apistat.domain, 'test')
        self.assertEqual(apistat.path, '/test/a/path/')
        self.assertIsNone(apistat.user)
        self.assertEqual(apistat.delay, 100)
        self.assertEqual(apistat.status, 200)

    def test_create_without_user(self):
        request = self.factory.get('/test/a/path/')
        delay = 100
        status = 200
        apistat = APIStat.objects.create_from_request(request=request, delay=delay, status=status)
        self.assertIsNotNone(apistat)
        self.assertEqual(apistat.domain, 'test')
        self.assertEqual(apistat.path, '/test/a/path/')
        self.assertIsNone(apistat.user)
        self.assertEqual(apistat.delay, 100)
        self.assertEqual(apistat.status, 200)

    def test_create_with_user(self):
        request = self.factory.get('/test/a/path/')
        request.user = self.user
        delay = 100
        status = 200
        apistat = APIStat.objects.create_from_request(request=request, delay=delay, status=status)
        self.assertIsNotNone(apistat)
        self.assertEqual(apistat.domain, 'test')
        self.assertEqual(apistat.path, '/test/a/path/')
        self.assertIsNotNone(apistat.user)
        self.assertEqual(apistat.delay, 100)
        self.assertEqual(apistat.status, 200)

    def test_count(self):
        request_a = self.factory.get('/test/a/path/')
        request_b = self.factory.get('/test/b/path/')
        request_c = self.factory.get('/another/path/')
        request_a.user = AnonymousUser()
        request_b.user = AnonymousUser()
        request_c.user = AnonymousUser()
        delay = 100
        status = 200
        apistat_a = APIStat.objects.create_from_request(request=request_a, delay=delay, status=status)
        APIStat.objects.create_from_request(request=request_b, delay=delay+100, status=status)
        APIStat.objects.create_from_request(request=request_c, delay=delay+200, status=status)
        self.assertEqual(apistat_a.total_count(), 3)
        self.assertEqual(apistat_a.domain_count(), 2)
        self.assertEqual(apistat_a.path_count(), 1)

    def test_max_delay(self):
        request_a = self.factory.get('/test/a/path/')
        request_b = self.factory.get('/test/b/path/')
        request_c = self.factory.get('/another/path/')
        request_a.user = AnonymousUser()
        request_b.user = AnonymousUser()
        request_c.user = AnonymousUser()
        delay = 100
        status = 200
        apistat_a = APIStat.objects.create_from_request(request=request_a, delay=delay, status=status)
        APIStat.objects.create_from_request(request=request_b, delay=delay + 100, status=status)
        APIStat.objects.create_from_request(request=request_c, delay=delay + 200, status=status)
        self.assertEqual(apistat_a.max_delay(), 300)
        self.assertEqual(apistat_a.max_domain_delay(), 200)
        self.assertEqual(apistat_a.max_path_delay(), 100)

    def test_avg_delay(self):
        request_a = self.factory.get('/test/a/path/')
        request_b = self.factory.get('/test/b/path/')
        request_c = self.factory.get('/another/path/')
        request_a.user = AnonymousUser()
        request_b.user = AnonymousUser()
        request_c.user = AnonymousUser()
        delay = 100
        status = 200
        apistat_a = APIStat.objects.create_from_request(request=request_a, delay=delay, status=status)
        APIStat.objects.create_from_request(request=request_b, delay=delay + 100, status=status)
        APIStat.objects.create_from_request(request=request_c, delay=delay + 200, status=status)
        self.assertEqual(apistat_a.avg_delay(), 200)
        self.assertEqual(apistat_a.avg_domain_delay(), 150)
        self.assertEqual(apistat_a.avg_path_delay(), 100)







# Generated by Django 3.1.1 on 2020-10-09 12:42

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('apistats', '0004_auto_20201009_1151'),
    ]

    operations = [
        migrations.AddField(
            model_name='apistat',
            name='user',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='apistat',
            name='delay',
            field=models.IntegerField(verbose_name='delay (ms) between request and response'),
        ),
        migrations.AlterField(
            model_name='apistat',
            name='domain',
            field=models.CharField(db_index=True, max_length=50, verbose_name='Application domain'),
        ),
        migrations.AlterField(
            model_name='apistat',
            name='ip',
            field=models.GenericIPAddressField(blank=True, null=True, verbose_name='Client IP address'),
        ),
        migrations.AlterField(
            model_name='apistat',
            name='method',
            field=models.CharField(max_length=10, verbose_name='call method'),
        ),
        migrations.AlterField(
            model_name='apistat',
            name='path',
            field=models.CharField(db_index=True, max_length=255, verbose_name='Application path'),
        ),
        migrations.AlterField(
            model_name='apistat',
            name='query',
            field=models.CharField(blank=True, max_length=2000, verbose_name='Query parameters'),
        ),
        migrations.AlterField(
            model_name='apistat',
            name='record_time',
            field=models.DateTimeField(auto_now=True, primary_key=True, serialize=False, verbose_name='Recorded at'),
        ),
    ]



# Generated by Django 3.1.1 on 2020-10-08 22:37

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('apistats', '0002_auto_20201008_2203'),
    ]

    operations = [
        migrations.AddField(
            model_name='apistat',
            name='domain',
            field=models.CharField(db_index=True, default='test', max_length=50),
            preserve_default=False,
        ),
    ]



# Generated by Django 3.1.1 on 2020-10-09 11:51

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('apistats', '0003_apistat_domain'),
    ]

    operations = [
        migrations.AddField(
            model_name='apistat',
            name='delay',
            field=models.IntegerField(default=0, verbose_name='delay between request and response'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='apistat',
            name='status',
            field=models.IntegerField(default=200, verbose_name='Response status code'),
            preserve_default=False,
        ),
    ]



from django.conf import settings
from django.contrib.auth.models import User
from django.db import models
from django.http import HttpRequest


class APIStat:
    pass


class APIStatManager(models.Manager):

    def create_from_request(
            self,
            request: HttpRequest,
            delay: int = 0,
            status: int = 200
    ) -> (APIStat, None):
        """
        Create a APIStat record in database and a APIStat object
        :param request: HTTPRequest
        :param delay: delay between request and response
        :param status: HTTP status of response
        :return:
        """
        excluded_domains = getattr(settings, 'APISTAT_EXCLUDED_DOMAINS', frozenset())
        domain = self._get_domain(request.path_info)
        if domain in excluded_domains:
            return
        apistat = APIStat()
        apistat.method = request.method
        apistat.query = request.META.get('QUERY_STRING')
        apistat.path = request.path_info
        apistat.domain = domain
        apistat.delay = delay
        apistat.status = status
        apistat.ip = request.headers.get('X-Real-IP', request.META.get('REMOTE_ADDR'))
        apistat.user = request.user if hasattr(request, 'user') and request.user.is_authenticated else None
        apistat.save()
        return apistat

    def _get_domain(self, path: str) -> str:
        """
        Extract first block from full path
        :param path: HTTPRequest path
        :return: first block of path (/test/hello/ => test)
        """
        split_path = path.split('/')
        if len(split_path) >= 2:
            return split_path[1]
        else:
            return ''


class APIStat(models.Model):
    """
    APIStat model class
    """
    record_time = models.DateTimeField("Recorded at", auto_now=True, primary_key=True)
    method = models.CharField("call method", max_length=10)
    domain = models.CharField("Application domain", max_length=50, db_index=True)
    path = models.CharField("Application path", max_length=255, db_index=True)
    query = models.CharField("Query parameters", max_length=2000, blank=True)
    ip = models.GenericIPAddressField("Client IP address", null=True, blank=True)
    delay = models.IntegerField("delay (ms) between request and response")
    status = models.IntegerField("Response status code")
    user = models.ForeignKey(User, null=True, blank=True, on_delete=models.PROTECT)
    objects = APIStatManager()

    @property
    def record_month(self) -> str:
        """
        Return YEAR-MONTH of record_time
        :return: YEAR-MONTH as string
        """
        return self.record_time.strftime('%Y-%M')

    @classmethod
    def total_count(cls) -> int:
        """
        Count total nubmer of requests
        :return: Number of requests
        """
        return cls.objects.count()

    def path_count(self) -> int:
        """
        Count number of requests for a given path
        :return: number of requests
        """
        return APIStat.objects.filter(path=self.path).count()

    def domain_count(self) -> int:
        """
        Count number of requests for a given domain
        :return: number of requests
        """
        return APIStat.objects.filter(domain=self.domain).count()

    @classmethod
    def avg_delay(cls):
        """
        Computes avg response delay for all responses
        :return: average response delay
        """
        return cls.objects.aggregate(avg_delay=models.Avg('delay')).get('avg_delay', 0)

    def avg_path_delay(self) -> float:
        """
        Computes avg response delay for a given path
        :return: average response delay
        """
        return APIStat.objects.filter(path=self.path).aggregate(avg_delay=models.Avg('delay')).get('avg_delay', 0)

    def avg_domain_delay(self) -> float:
        """
        Computes avg response delay for a given domain
        :return: average response delay
        """
        return APIStat.objects.filter(domain=self.domain).aggregate(avg_delay=models.Avg('delay')).get('avg_delay', 0)

    @classmethod
    def max_delay(cls) -> float:
        """
        Computes max response delay for all responses
        :return: max response delay
        """
        return cls.objects.aggregate(max_delay=models.Max('delay')).get('max_delay', 0)

    def max_path_delay(self) -> float:
        """
        Computes max response delay for a given path
        :return: max response delay
        """
        return APIStat.objects.filter(path=self.path).aggregate(max_delay=models.Max('delay')).get('max_delay', 0)

    def max_domain_delay(self):
        """
        Computes max response delay for a given domain
        :return: max response delay
        """
        return APIStat.objects.filter(domain=self.domain).aggregate(max_delay=models.Max('delay')).get('max_delay', 0)

    class Meta:
        ordering = ['-record_time', 'path']
        get_latest_by = ['-record_time', 'path']



from time import time

from django.http import HttpRequest, HttpResponse

from .models import APIStat


class LoggingMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response
        # One-time configuration and initialization.

    def __call__(self, request: HttpRequest) -> HttpResponse:
        """
        Call middleware
        :param request: HTTPRequest
        :return: HTTPResponse
        """
        start_time = time()
        response = self.get_response(request)

        # Code to be executed for each request/response after
        # the view is called.
        delay = time() - start_time
        APIStat.objects.create_from_request(
            request=request,
            delay=delay * 1000,
            status=response.status_code
        )
        return response
